#include "m_pd.h"
#include <stdlib.h>
#include <math.h>
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

/* ------------------------ dyna~ ----------------------------- */

/* tilde object to adjust an rms value based on attack and decay */

static t_class *dyna_class;

typedef struct _dyna
{
    t_object x_obj;			// obligatory variable
    t_float x_f;			// stores message recieved on audio inlet
	t_float samplerate;		// sample rate

	t_float threshold;		// threshold in rms
	t_float thresholddb;	// threshold in dB
	t_int trigger;			// if the signal has passed the threshold
	t_int status;			// 0 = no adjustment, 1 = attacking, 2 = decaying

	t_int attack;			// attack in samples
	t_int attackms;			// attack in milliseconds
	t_int asample;			// amount of samples after attack trigger

	t_int decay;			// decay in samples
	t_int decayms;			// decay in milliseconds
	t_int dsample;			// amount of samples after decay trigger
} t_dyna;

static t_int *dyna_perform(t_int *w)
{
	t_dyna *x = (t_dyna *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);

	// number of samples passed
	int blocksize = n;
	// sample from 0 up
	int sample = 0;
	// converting from dB to rms
	x->threshold = (t_float)pow((double)10, (double)(x->thresholddb/20))/100000;
	// converting from milliseconds to samples
	x->attack = x->attackms * 0.001f * x->samplerate;
	x->decay = x->decayms * 0.001f * x->samplerate;

    while (n--)
    {

		if( *(in+sample) > x->threshold && x->trigger == 0)
		{
			x->trigger = 1;		// set to being above threshold
			x->status = 1;		// set status to attacking
			x->asample = 0;		// reset attack sample
		}
		else if( *(in+sample) < x->threshold && x->trigger == 1)
		{
			x->trigger = 0;		// set to being below threshold
			x->status = 2;		// set status to decaying
			x->dsample = 0;		// reset decay sample
		}

		if( x->status == 1 )
		{
			// run attacking algorithm
			*(out+sample) = (*(in+sample) - x->threshold) * x->asample/x->attack + x->threshold;
			// turn off attacking status
			if( x->asample == x->attack )
			{
				x->status = 0;
			}
			x->asample++;
		}
		else if( x->status == 2 )
		{
			// run decaying algorithm
			*(out+sample) = (*(in+sample) - x->threshold) * x->dsample/x->decay + x->threshold;
			// turn off decaying status
			if( x->dsample == x->decay )
			{
				x->status = 0;
			}
			x->dsample++;
		}
		// output actual rms adjustment
		else
		{
			*(out+sample) = *(in+sample);
		}
		sample++;
    }
    return (w+5);
}

static void dyna_dsp(t_dyna *x, t_signal **sp)
{
	x->samplerate = sp[0]->s_sr;
    dsp_add(dyna_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

// setting threshold
void dyna_threshset(t_dyna *x, t_floatarg f)
{
	if( f >= 1 && f <= 100 )
		x->thresholddb = f;
	else if( f < 1 )
		x->thresholddb = 1;
	else if( f > 100 )
		x->thresholddb = 100;
}

// setting attack time
void dyna_attackset(t_dyna *x, t_floatarg f)
{
	if( f >= 1 && f <= 1000 )
		x->attackms = f;
	else if( f < 1 )
		x->attackms = 1;
	else if( f > 1000 )
		x->attackms = 1000;
}

// setting decay time
void dyna_decayset(t_dyna *x, t_floatarg f)
{
	if( f >= 1 && f <= 1000 )
		x->decayms = f;
	else if( f < 1 )
		x->decayms = 1;
	else if( f > 1000 )
		x->decayms = 1000;
}

static void *dyna_new(t_floatarg f, t_floatarg g, t_floatarg h)
{
    t_dyna *x = (t_dyna *)pd_new(dyna_class);
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("thresh"));
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("attack"));
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("decay"));
    outlet_new(&x->x_obj, gensym("signal"));
    x->x_f = 0;
	x->trigger = 0;
	x->status = 0;
	x->asample = 0;
	x->dsample = 0;

	// setting initialized values
	dyna_threshset(x, f);
	dyna_attackset(x, g);
	dyna_decayset(x, h);

    return (x);
}

void dyna_tilde_setup(void)
{
    dyna_class = class_new(gensym("dyna~"), (t_newmethod)dyna_new, 0,
    	sizeof(t_dyna), 0, A_DEFFLOAT, A_DEFFLOAT, A_DEFFLOAT, 0);
    CLASS_MAINSIGNALIN(dyna_class, t_dyna, x_f);
	class_addmethod(dyna_class, (t_method)dyna_threshset, gensym("thresh"), A_FLOAT, 0);
	class_addmethod(dyna_class, (t_method)dyna_attackset, gensym("attack"), A_FLOAT, 0);
	class_addmethod(dyna_class, (t_method)dyna_decayset, gensym("decay"), A_FLOAT, 0);
    class_addmethod(dyna_class, (t_method)dyna_dsp, gensym("dsp"), (t_atomtype)0);
}